/* This file is part of Idest.
   Copyright (C) 2009 Sergey Poznyakoff
 
   Idest is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.
 
   Idest is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with Idest.  If not, see <http://www.gnu.org/licenses/>. */

#include "idest.h"

static char *id3v1_genre[] = {
	"Blues",
        "Classic Rock",
        "Country",
        "Dance",
        "Disco",
        "Funk",
        "Grunge",
        "Hip-Hop",
        "Jazz",
        "Metal",
        "New Age",
        "Oldies",
        "Other",
        "Pop",
        "R&B",
        "Rap",
        "Reggae",
        "Rock",
        "Techno",
        "Industrial",
        "Alternative",
        "Ska",
        "Death Metal",
        "Pranks",
        "Soundtrack",
        "Euro-Techno",
        "Ambient",
        "Trip-Hop",
        "Vocal",
        "Jazz+Funk",
        "Fusion",
        "Trance",
        "Classical",
        "Instrumental",
        "Acid",
        "House",
        "Game",
        "Sound Clip",
        "Gospel",
        "Noise",
        "AlternRock",
        "Bass",
        "Soul",
        "Punk",
        "Space",
        "Meditative",
        "Instrumental Pop",
        "Instrumental Rock",
        "Ethnic",
        "Gothic",
        "Darkwave",
        "Techno-Industrial",
        "Electronic",
        "Pop-Folk",
        "Eurodance",
        "Dream",
        "Southern Rock",
        "Comedy",
        "Cult",
        "Gangsta",
        "Top 40",
        "Christian Rap",
        "Pop/Funk",
        "Jungle",
        "Native American",
        "Cabaret",
        "New Wave",
        "Psychadelic",
        "Rave",
        "Showtunes",
        "Trailer",
        "Lo-Fi",
        "Tribal",
        "Acid Punk",
        "Acid Jazz",
        "Polka",
        "Retro",
        "Musical",
        "Rock & Roll",
        "Hard Rock",
        "Folk",
        "Folk-Rock",
        "National Folk",
        "Swing",
        "Fast Fusion",
        "Bebob",
        "Latin",
        "Revival",
        "Celtic",
        "Bluegrass",
        "Avantgarde",
        "Gothic Rock",
        "Progressive Rock",
        "Psychedelic Rock",
        "Symphonic Rock",
        "Slow Rock",
        "Big Band",
        "Chorus",
        "Easy Listening",
        "Acoustic",
        "Humour",
        "Speech",
        "Chanson",
        "Opera",
        "Chamber Music",
        "Sonata",
        "Symphony",
        "Booty Bass",
        "Primus",
        "Porn Groove",
        "Satire",
        "Slow Jam",
        "Club",
        "Tango",
        "Samba",
        "Folklore",
        "Ballad",
        "Power Ballad",
        "Rhythmic Soul",
        "Freestyle",
        "Duet",
        "Punk Rock",
        "Drum Solo",
        "A capella",
        "Euro-House",
        "Dance Hall"
};

size_t id3v1_genre_count = sizeof(id3v1_genre)/sizeof(id3v1_genre[0]);

struct id3v1_tab {
	char *name;
	int ind;
	size_t len;
};

struct id3v1_tab id3v1_tab[128];

static int
id3v1_tab_cmp(const void *a, const void *b)
{
	const struct id3v1_tab *ap = a;
	const struct id3v1_tab *bp = b;
	return strcasecmp(ap->name, bp->name);
}

static int
id3v1_fill_tab()
{
	int i;

	for (i = 0; i < id3v1_genre_count; i++) {
		id3v1_tab[i].name = id3v1_genre[i];
		id3v1_tab[i].ind = i;
		id3v1_tab[i].len = strlen(id3v1_tab[i].name);
	}
	qsort(id3v1_tab, id3v1_genre_count, sizeof(id3v1_tab[0]),
	      id3v1_tab_cmp);
}

const char *
id3v1_n_to_genre(unsigned n)
{
	if (n < id3v1_genre_count)
		return id3v1_genre[n];
	return "Unknown";
}

int
id3v1_genre_to_n(const char *genre)
{
	int i, index = -1;
	int ambiguous;
	size_t glen = strlen(genre);
	if (!id3v1_tab[0].name)
		id3v1_fill_tab();		
	for (i = 0; i < id3v1_genre_count; i++) {
		if (glen <= id3v1_tab[i].len
		    && strncasecmp(id3v1_tab[i].name, genre, glen) == 0) {
			if (glen == id3v1_tab[i].len) 
				return id3v1_tab[i].ind;
			else if (index == -1)
				index = id3v1_tab[i].ind;
			else
				return -2;
		} else if (index != -1 && genre[0] != id3v1_tab[i].name[0])
			break;
	}
	return index;
}
		
void
id3v1_copy_string(const char *name, char *ptr, size_t size, const char *arg)
{
	size_t len = strlen(arg);
	if (len > size) {
		error(0, 0, "%s truncated", name);
		len = size;
	}
	memcpy(ptr, arg, len);
}
		
int
id3v1_read(FILE *fp, struct id3v1_block *blk)
{
	struct id3v1_block id;
	if (fseek(fp, -sizeof(id), SEEK_END))
		return -1;
	if (fread(&id, sizeof(id), 1, fp) != 1) 
		return -1;
	if (memcmp(id.header, ID3V1_TAG, ID3V1_HEADER_LEN))
		return 1;
	memcpy(blk, &id, sizeof(blk[0]));
	return 0;
}

int
id3v1_write(FILE *fp, struct id3v1_block *blk)
{
	struct id3v1_block id;
	long offset;
	
	switch (id3v1_read(fp, &id)) {
	case 0:
		offset = -sizeof(id);
		break;
	case 1:
		offset = 0;
		break;
	default:
		return -1;
	}
	memcpy(blk, ID3V1_TAG, ID3V1_HEADER_LEN);
	if (fseek(fp, offset, SEEK_END))
		return -1;
	if (fwrite(blk, sizeof(blk[0]), 1, fp) != 1) 
		return -1;
	return 0;
}

#define ID3_MERGE_FIELD(new,old,fld) do {				\
		if (new->fld[0] == 0 && old->fld)			\
			memcpy(new->fld, old->fld, sizeof(new->fld));	\
	} while (0)

void
id3v1_merge(struct id3v1_block *id, const struct id3v1_block *old)
{
	ID3_MERGE_FIELD(id,old,title);
	ID3_MERGE_FIELD(id,old,artist);
	ID3_MERGE_FIELD(id,old,album);
	ID3_MERGE_FIELD(id,old,year);
	ID3_MERGE_FIELD(id,old,comment);
}
		
void
id3v1_init(struct id3v1_block *blk)
{
	memset(blk, 0, sizeof(blk[0]));
	memcpy(blk, ID3V1_TAG, ID3V1_HEADER_LEN);
}


void
set_id3v1(const char *name, const struct id3v1_block *blk)
{
	int rc;
	struct id3v1_block new, old;
	FILE *fp = fopen(name, "r+");
	if (!fp) 
		error(1, errno, "cannot open file %s", name);

	verify_mp3(fp, name);
	memcpy(&new, blk, sizeof(new));
	switch (id3v1_read(fp, &old)) {
	case 0:
		id3v1_merge(&new, &old);
		break;
	case 1:
		break;
	default:
		error(1, errno, "%s: read error", name);
	}
	if (id3v1_write(fp, &new))
		error(1, errno, "%s: write error", name);
	
	fclose(fp);		
}

void
query_id3v1(const char *name)
{
	struct id3v1_block id;
	FILE *fp = fopen(name, "r");
	int rc;
	if (!fp) 
		error(1, errno, "cannot open file %s", name);

	verify_mp3(fp, name);
	rc = id3v1_read(fp, &id);
	fclose(fp);
	
	switch (rc) {
	case 0:
		printf("title: %-.30s\n", id.title);
		printf("artist: %-.30s\n", id.artist);
		printf("album: %-.30s\n", id.album);
		printf("year: %-.4s\n", id.year);
		printf("comment: %-.28s\n", id.comment);
		printf("genre: %s\n", id3v1_n_to_genre(id.genre[0]));
		if (id.pad[0] == 0)
			printf("track: %d\n", id.track[0]);
		break;
	case 1:
		break;
	case -1:
		error(1, errno, "%s: read error", name);
	}
}

void
del_id3v1(const char *name)
{
	error(1, 0, "Delete is not implemented");
}



